//===========================================================================
// PPKG Puppy-Package-Handler
//===========================================================================
//
// Copyright (C) 2009 Wolfram Schuler Nuremberg, Germany.
// Written by Wolfram Schuler <wosh@online.de>.
// 
// This file is the source code of PPKG.
// 
// PPKG is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
// 
// PPKG is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with PPKG -- see the file COPYING.
// If not, see <http://www.gnu.org/licenses/>.
//===========================================================================
//Notes
//
//PPKG has been written to overcome performance problems in the Puppy Linux
//package management.
//
//It wants to be an example how software can be written in a professional
//approach. As it uses some experienced software techniques it might be
//regarded as a source of inspiration for software solutions written in C.
//
//I hope that I have included enough subtleties and specialties to please
//beginners and experienced programmers. Although I have protected the whole
//because it might become a part of the Puppy Linux package management, I
//encourage you to reenact and to interalize those parts that you did not
//know and apply yet.
//
//This software was written claiming to become the fastest Package Info
//converter that can be realized in C. I personally have pulled out all the
//stops and reached my limits to make this software as effective as possible.
//I can not do it better, but maybe you.

//So I have made this software part of a contest.
//Details about the conditions can be found in the Puppy Linux Discussion.
//Forum http://wwww.murga-linux.com/puppy
//Topic:http://www.murga-linux.com/puppy/viewtopic.php?t=47763&sid=bce595c8184d849fe62a61a46453c245
//
//Is there anybody out there who can beat me?
//
//October 2009 Wosh
//===========================================================================

//if the project becomes larger this sould become a separate module
//===========================================================================
// PKGERR.C - Puppy-Package-Handler Error Functions
//===========================================================================
#include <errno.h>			//errno
#include <stdio.h>			//fprintf
#include <stdlib.h>			//exit malloc
#include <string.h>			//strchr strerror

#include "pkgglb.h"

//===========================================================================
//LOCAL DATA
//===========================================================================
static const uchar* err_program_p;	//program-name pointer

//===========================================================================
//GLOBAL FUNCTIONS
//===========================================================================

//===========================================================================
void ERR_ErrorFatal(			//error-fatal
  const uchar* const	text_p)		//in: text pointer
{
  fprintf(stderr,
    "%s: Error: %s\n",
    err_program_p, text_p);

  exit(1);				//exit with error
}

//===========================================================================
void ERR_ErrorFatalLine(		//error-fatal-line
  const uchar* const	text_p,		//in: text pointer
  const uchar* const	fileName_p,	//in: file-name pointer
  const ulong		lineNr)		//in: line-number
{
  fprintf(stderr,
    "%s: Fatal error %s %lu: %s\n",
    err_program_p, fileName_p, lineNr, text_p);

  exit(1);				//exit with error
}

//===========================================================================
void ERR_ErrorFatalSystem(		//error-fatal-system
  const uchar* const	text_p,		//in: text pointer
  const uchar* const	fileName_p)	//in: file-name pointer
{
  fprintf(stderr,
    "%s: File %s: %s. Error: %s",
    err_program_p, fileName_p, text_p, strerror(errno));

  exit(1);				//exit with error
}

//===========================================================================
void ERR_Init(				//error-initialisation
  const uchar* const program_p)		//in: program-name pointer
{
  err_program_p = program_p;		//program-name pointer
}

//if the project becomes larger this sould become a separate module
//===========================================================================
// PKGMEM.C - Puppy-Package-Handler-Memory Functions
//===========================================================================
#include <string.h>			//memcpy

//===========================================================================
//GLOBAL NAMED-CONSTANTS
//===========================================================================
#define MEM_MEM_SIZE_MAX ((MEM_SIZE_T) 0xFFFFFFE0ul) //memory-size-maximum

//===========================================================================
//GLOBAL TYPES
//===========================================================================
typedef ulong	MEM_SIZE_T;		//memory-size type

//===========================================================================
//GLOBAL FUNCTIONS
//===========================================================================
//All MEM-functions actually should be written in assembler as they are
//crucial for the performance of a program

//===========================================================================
uchar* MEM_Copy(			//memory-copy
  const MEM_SIZE_T	count,		//in: count
  const uchar* const	source_p,	//in: source pointer
  uchar* const		target_p)	//in/out: target pointer
					//out: target-end pointer
//this function is wrapper around memcpy because memcpy misses the feature
//to return the target-end pointer which is usefull for concatenations.
{
  memcpy(				//memory-copy
    target_p,				//in/out: target pointer
    source_p,				//in: source pointer
    count);				//in: count

  return &target_p[count];		//target-end pointer
}

//===========================================================================
uchar* MEM_CharFromGroup(		//memory-character-from-group
  const uchar* const	group_p,	//in: group pointer
  const MEM_SIZE_T	count,		//in: count
  const uchar* const	mem_p)		//in: memory pointer
					//out: character pointer
//return a pointer if a character from a group of characters is found
//or NULL if none of the characters was found
{
  MEM_SIZE_T	i;			//index
  uchar*	memChar_p;		//memory-character pointer

  i = count;				//index
  memChar_p = (uchar*) mem_p;		//memory-character pointer

  while (i--)				//for whole memory block
  {
    const uchar* groupChar_p;		//group-character pointer

    //for whole group
    for (groupChar_p = group_p; *groupChar_p != 0; groupChar_p++)
      if (*memChar_p == *groupChar_p)	//character found?
	return memChar_p;		//memory-character pointer

    memChar_p++;			//memory-character pointer
  }
  return GLB_NULL;			//character not found
}

//if the project becomes larger this sould become a separate module
//===========================================================================
// PKGFIO.C - Puppy-Package-Handler File-Input-Output
//===========================================================================
#include <fcntl.h>			//O_xxx flags
#include <sys/stat.h>			//stat S_IREAD S_IWRITE
#include <unistd.h>			//read

//===========================================================================
//GLOBAL TYPES
//===========================================================================
typedef	size_t	FIO_SIZE_T;		//file-size type

typedef enum
{
  FIO_FLR_SUCCESS =	0,		//success
  FIO_FLR_ERR_OPEN =	1,		//open-error
  FIO_FLR_ERR_SIZE =	2,		//size-error
  FIO_FLR_ERR_MEMORY =	3,		//memory-allocation-error
  FIO_FLR_ERR_READ =	4,		//read-error
} FIO_FILE_RSL_ET;			//file-result enum-type

typedef enum
{
  FIO_OPT_NONE =	0,		//none
  FIO_OPT_EXIT_ON_ERR =	1 << 0,		//exit-on-error
  FIO_OPT_OUTPUT_ERR =	1 << 1,		//output-error
  FIO_OPT_EOS_0 =	1 << 2,		//end-of-string-0
} FIO_OPTION_ET;			//option enum-type

//===========================================================================
//GLOBAL FUNCTIONS
//===========================================================================

//===========================================================================
GLB_RESULT_T FIO_FileSize(		//file-size
  const int		file_h,		//in: file handle
  ulong* const		fileSize_p)	//in/out: file-size pointer
					//out: valid-size-found
//get a file's size
{
  struct stat	fileStatus;		//file-status

  if (fstat(				//file-status
    file_h,				//in: file handle
    &fileStatus)			//in/out: file-status pointer
					//out: result
    == -1)				//failure?
  {
    *fileSize_p = 0;			//file-size pointer
    return GLB_RSL_FAILURE;		//failure
  }
  *fileSize_p = (FIO_SIZE_T) fileStatus.st_size; //file-size pointer
  return GLB_RSL_SUCCESS;		//success
}

//===========================================================================
FIO_FILE_RSL_ET FIO_FileToMemory(	//file-to-memory
  const uchar* const fileName_p,	//in: file-name pointer
  const FIO_OPTION_ET	option,		//in: option
  uchar** const		mem_pp,		//in/out: memory pointer-pointer
  MEM_SIZE_T* const	memSize_p)	//in/out: memory-size pointer
					//out: file-result
//write a whole file to memory and do some tedious error handling
{
  FIO_FILE_RSL_ET fileResult;		//file-result
  int		file_h;			//file-handle
  ulong		fileSize;		//file-size

  fileResult = FIO_FLR_SUCCESS;		//success

  if ((file_h = open(			//open
    (char*) fileName_p,			//in: file-name pointer
    O_RDONLY))				//in: openKind
					//out: file handle
    <= 0)				//failure?
    fileResult = FIO_FLR_ERR_OPEN;	//open-error
  else
  {
    if (FIO_FileSize(			//file-size
	file_h,				//in: file handle
	&fileSize)			//in/out: file-size pointer
					//out: size-valid
	== GLB_RSL_FAILURE)		//failure?
      fileResult = FIO_FLR_ERR_SIZE;	//file-size-error
    else
    {
      FIO_SIZE_T bytesRead;		//bytes-read-from-file
      MEM_SIZE_T memSize;		//memory-size

      memSize = fileSize;		//memory-size

      if ((option & FIO_OPT_EOS_0) != 0) //end-of-string-0?
	memSize++;			//memory-size

      if (fileSize > MEM_MEM_SIZE_MAX)	//or file-size to big?
      {
	errno = ENOMEM;
	fileResult = FIO_FLR_ERR_SIZE;	//file-size-error
      }
      else if ((*mem_pp = malloc(	//memory-allocate
	memSize))			//in: size
					//out: memory pointer
	== GLB_NULL)			//invalid memory pointer?
      {
	errno = ENOMEM;
	fileResult = FIO_FLR_ERR_MEMORY; //memory-allocation-error
      }
      else
      {
	bytesRead = read(		//read
	  file_h,			//in: file handle
	  *mem_pp,			//in/out: data pointer
	  (unsigned) fileSize);		//in: count
					//out: number-read

	if (bytesRead <= 0)		//number-read error?
	  fileResult = FIO_FLR_ERR_READ; //read-error
	else if (bytesRead != (FIO_SIZE_T) fileSize) //not read completely?
	{
	  errno = EILSEQ;
	  fileResult = FIO_FLR_ERR_READ; //read-error
	}
	else
	{
	  if ((option & FIO_OPT_EOS_0) != 0) //end-of-string-0?
	    (*mem_pp)[memSize - 1] = '\0'; //end-of-string

	  if (memSize_p != GLB_NULL)	//valid memory-size pointer?
	    *memSize_p = memSize;	//memory-size pointer
	}
      }
    }
    close(				//file-close
      file_h);				//in: file handle
					//out: close-result
  }
  if (fileResult != FIO_FLR_SUCCESS)	//no success?
  {
    if ((option & FIO_OPT_OUTPUT_ERR) != 0) //output-error?
      ERR_ErrorFatalSystem(		//error-fatal-system
	(uchar*) "Reading to memory",	//in: text pointer
	fileName_p);			//in: file-name pointer

    if ((option & FIO_OPT_EXIT_ON_ERR) != 0) //exit-on-error?
      exit(fileResult);			//exit with error
  }
  return fileResult;			//file-result
}

//===========================================================================
// PPKG.C - Puppy-Package-Handler
//===========================================================================
#include <alloca.h>			//alloca
#include <ctype.h>			//isspace

//GNU extension of glibc
extern void* memrchr(const void* mem_p, int c, size_t size);

//===========================================================================
//LOCAL NAMED-CONSTANTS
//===========================================================================
//#define pkg_POLITICAL_CORRECT		//avoid dubious fast constructs
#define pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer

#define pkg_LINE_MAX		4096	//line-size-maximum
#define pkg_OUTPUT_LINE_COUNT	32	//output-buffer-line-count
#define pkg_SEPARATOR		'|'	//separator

//===========================================================================
//LOCAL MACROS
//===========================================================================
#define pkg_String_M(string_s)\
  {sizeof(string_s) - 1, (uchar*) string_s} //string macro

//===========================================================================
//LOCAL TYPES
//===========================================================================
typedef enum
{
  pkg_ICT_BUILDINGBLOCK	= 0,		//buildingblock
  pkg_ICT_CALCULATE	= 1,		//calculate
  pkg_ICT_DESKTOP	= 2,		//desktop
  pkg_ICT_DEVELOP	= 3,		//develop
  pkg_ICT_DOCUMENT	= 4,		//document
  pkg_ICT_FILESYSTEM	= 5,		//filesystem
  pkg_ICT_FUN		= 6,		//fun
  pkg_ICT_GRAPHIC	= 7,		//graphic
  pkg_ICT_HELP		= 8,		//help
  pkg_ICT_INTERNET	= 9,		//internet
  pkg_ICT_MULTIMEDIA	= 10,		//multimedia
  pkg_ICT_NETWORK	= 11,		//network
  pkg_ICT_PERSONAL	= 12,		//personal
  pkg_ICT_SYSTEM	= 13,		//system
  pkg_ICT_UTILITY	= 14,		//utility
  pkg_ICT_COUNT		= 15,		//count
} pkg_I_CATEGORY_ET;			//category-index enum-type

typedef enum
{
  //puppy data-base row
  pkg_PKI_CUT_BASE	= 0,		//cut-file-base
  pkg_PKI_PACKAGE	= 1,		//package
  pkg_PKI_VERSION	= 2,		//version
  pkg_PKI_RELEASE	= 3,		//version
  pkg_PKI_CATEGORY	= 4,		//category
  pkg_PKI_SIZE		= 5,		//size
  pkg_PKI_DIRECTORY	= 6,		//directory
  pkg_PKI_FILE		= 7,		//file
  pkg_PKI_DEPS		= 8,		//dependencies
  pkg_PKI_DESCRIPTION	= 9,		//description
  pkg_PKI_ROW_COUNT	= 10,		//row-count

  //internal items
  pkg_PKI_NEXT_PACKAGE	= 10,		//next-package
  pkg_PKI_COUNT		= 11,		//count
} pkg_ITEM_ET;				//puppy-package-item enum-type

typedef struct
{
  MEM_SIZE_T	length;			//length
  uchar*	string_p;		//string pointer
} pkg_STR_T;				//string type

typedef struct
{
  pkg_STR_T	name;			//name pointer
  pkg_I_CATEGORY_ET iCategory;		//puppy-category-index
} pkg_CTG_PACKAGE_T;			//category-package type

typedef struct
{
  const uchar*	fileName_p;		//file-name pointer
  MEM_SIZE_T	textLength;		//text-length
  uchar*	text_bp;		//text buffer-pointer
  MEM_SIZE_T	packageCount;		//package-count
  pkg_CTG_PACKAGE_T* package_bp;	//package buffer-pointer
} pkg_CATEGORY_T;			//category type

typedef struct
{
  const uchar*	fileName_p;		//file-name pointer
  int		file_h;			//file-handle
  uchar		text_b[pkg_LINE_MAX *
    pkg_OUTPUT_LINE_COUNT];		//text buffer
  uchar*	line_p;			//line pointer
} pkg_OUTPUT_T;				//output type

typedef struct
{
  GLB_BOOL_T	excludeDebug_F;		//exclude-debug flag
  const uchar*	inputFileName_p;	//input-file-name pointer
  pkg_CATEGORY_T category;		//category
  pkg_OUTPUT_T	output;			//output
} pkg_CONVERT_T;			//convert type

typedef struct
{
  uchar*	start_p;		//start pointer
  uchar*	end_p;			//start pointer
  MEM_SIZE_T	nr;			//number
  GLB_BOOL_T	eot_F;			//end-of-text flag
} pkg_LINE_T;				//line type

typedef struct
{
  MEM_SIZE_T	lineNr;			//line-number
  pkg_STR_T	info_b[pkg_PKI_COUNT];	//info buffer
} pkg_PACKAGE_T;			//package type

typedef int (*pkg_COMPARE_FP_T)(	//compare-function-pointer type
  const void*		key_p,		//in: key pointer
  const void*		item_p);	//in: compare-item pointer
					//out: comparision-result

//===========================================================================
//LOCAL FUNCTION-DECLARATIONS
//===========================================================================

static void pkg_DebianConvertInfo(	//debian-convert-package-info
  pkg_CONVERT_T* const convert_p,	//in/out: convert pointer
  const pkg_PACKAGE_T* const package_p); //in: package pointer

static GLB_RESULT_T pkg_DebianGetIdText( //debian-get-id-text
  const pkg_STR_T	itemId,		//in: item-identifier
  const uchar* const	lineStart_p,	//in: line-start pointer
  const uchar* const	lineEnd_p,	//in: line-end pointer
  pkg_STR_T* const	text_p);	//in/out: text pointer
					//out: id-found

void pkg_SortedInsert(			//sorted-insert
  const MEM_SIZE_T	itemSize,	//in: item-size
  const void* const	item_p,		//in: item pointer
  const pkg_COMPARE_FP_T compare_fp,	//in: compare function-pointer
  const MEM_SIZE_T	itemCount,	//in: list-item-count
  void*			list_p);	//in/out: list pointer

static int pkg_StringCompare(		//string-compare
  const pkg_STR_T* const key_p,		//in: key pointer
  const pkg_STR_T* const listItem_p);	//in: list-item pointer
					//out: comparision-result

//===========================================================================
//LOCAL DATA
//===========================================================================

static const pkg_STR_T pkg_puppyCategory_a[] =	//puppy-category array
{
  pkg_String_M("BuildingBlock"),
  pkg_String_M("Calculate"),
  pkg_String_M("Desktop"),
  pkg_String_M("Develop"),
  pkg_String_M("Document"),
  pkg_String_M("Filesystem"),
  pkg_String_M("Fun"),
  pkg_String_M("Graphic"),
  pkg_String_M("Help"),
  pkg_String_M("Internet"),
  pkg_String_M("Multimedia"),
  pkg_String_M("Network"),
  pkg_String_M("Personal"),
  pkg_String_M("System"),
  pkg_String_M("Utility"),
};

//===========================================================================
//LOCAL FUNCTIONS
//===========================================================================

//===========================================================================
static int pkg_CStringItemCompare(	//c-string-item-compare
  const char* const	key_p,		//in: key pointer
  const char* const* const arrayItem_pp) //in: array-item pointer-pointer
					//out: comparision-result
//compare a c-string with a c-string pointer-array
{
  return strcmp(			//string-compare
    key_p,				//in: string-1 pointer
    *arrayItem_pp);			//in: string-2 pointer
					//out: comparision-result
}

//===========================================================================
static void pkg_DebianConvert(		//debian-convert
  pkg_CONVERT_T* const convert_p,	//in/out: convert pointer
  const MEM_SIZE_T	textLength,	//in: text-length
  const uchar* const	text_bp)	//in: text buffer-pointer
//convert Debian package infos to Puppy Linux package database rows
{
  #define PACKAGE_LENGTH\
    package.info_b[pkg_PKI_PACKAGE].length //package-length
  #define PACKAGE_P\
    package.info_b[pkg_PKI_PACKAGE].string_p //package pointer

  typedef enum
  {
    IDI_PACKAGE	= 5,			//package
  } DEBIAN_ID_INDEX_ET;			//debian-id-index enum-type

  typedef struct
  {
    pkg_STR_T	name;			//name
    pkg_ITEM_ET	iPuppy;			//puppy-data-base-item-index
  } DEBIAN_ID_T;			//debian-identifier type

  static const DEBIAN_ID_T dbc_debianId_a[] = //debian-identifier array
  {
    //name				puppy-data-base-item-index
    {pkg_String_M("Installed-Size:"),	pkg_PKI_SIZE},
    {pkg_String_M("Version:"),		pkg_PKI_VERSION},
    {pkg_String_M("Depends:"),		pkg_PKI_DEPS},
    {pkg_String_M("Filename:"),		pkg_PKI_DIRECTORY},
    {pkg_String_M("Description:"),	pkg_PKI_DESCRIPTION},
    {pkg_String_M("Package:"),		pkg_PKI_NEXT_PACKAGE},
  };

  GLB_BOOL_T	convert_F;		//convert flag
  GLB_BOOL_T	excludeDebug_F;		//exclude-debug flag
  pkg_LINE_T	line;			//line
  pkg_PACKAGE_T	package;		//package
  uchar		iStart;			//start-index
  uchar*	outputText_bp;		//output-text buffer-pointer

  if (convert_p->category.fileName_p !=
    GLB_NULL)				//valid category-file-name pointer?
  {
    #define MEAN_PKG_NAME_SIZE	(6 + 1)	//mean-package-name-size including LF
    #define PKG_EXTRA_COUNT	16	//extra-count

    MEM_SIZE_T	itemMax;		//item-maximum
    MEM_SIZE_T	packageCount;		//package-count
    pkg_CTG_PACKAGE_T package;		//package
    pkg_CTG_PACKAGE_T* package_bp;	//package buffer-pointer
    uchar*	ctgText_bp;		//category-text buffer-pointer

    //evaluate category file
    FIO_FileToMemory(			//file-to-memory
      convert_p->category.fileName_p,	//in: file-name pointer
      FIO_OPT_EOS_0 | FIO_OPT_OUTPUT_ERR | FIO_OPT_EXIT_ON_ERR, //in: option
      &convert_p->category.text_bp,	//in/out: memory pointer-pointer
      &convert_p->category.textLength);	//in/out: memory-size pointer
					//out: file-result

    //lazy but fast: avoid counting the packages within the text
    //allocation is done with some assumptions:
    //1. As one allocated item only consists of length-string pointer and
    //   category-index there is not much memory to allocate
    //2. Category files tend to be rather small, so it doesn't matter to
    //   waste some allocated stack memory.
    //3. The mean package-name length exceeds the size of 6 bytes.
    //   For the case the file is very small, provide some extra space.

    itemMax = (convert_p->category.textLength /
      MEAN_PKG_NAME_SIZE) + PKG_EXTRA_COUNT; //item-maximum

    if ((package_bp =
      convert_p->category.package_bp = alloca( //allocate-stack-memory
	itemMax * sizeof(pkg_CTG_PACKAGE_T))) //in: size
					//out: buffer pointer
      == GLB_NULL)			//invalid buffer pointer?
      ERR_ErrorFatal(			//error-fatal
	(uchar*) "Unable to allocate memory for categories");
					//in: text pointer

    //for debugging only
    //memset(package_bp, 0,
      //itemMax * sizeof(pkg_CTG_PACKAGE_T));

    package.iCategory = pkg_ICT_BUILDINGBLOCK; //category-index
    packageCount = 0;			//package-count
    convert_p->output.line_p =
      convert_p->output.text_b;		//output-line pointer
    ctgText_bp = convert_p->category.text_bp; //category-text buffer-pointer
    line.nr = 0;			//line-number
    line.eot_F = GLB_BOO_FALSE;		//end-of-text flag
    line.start_p = ctgText_bp;		//line-start pointer
    goto DCN2_L;			//skip the copying of the line-end

    do
    {
      //next line, skip line-feed
      line.start_p = &line.end_p[1];	//line-start pointer

      DCN2_L:
      //search line-end
      if ((line.end_p = (uchar*) strchr((char*) line.start_p, '\n'))
	== GLB_NULL)			//last-line?
      {
	if ((line.end_p = &ctgText_bp[
	  convert_p->category.textLength - 1])
	  == line.start_p)		//line-end and line-start same?
	  break;			//nothing more to do

	line.eot_F = GLB_BOO_TRUE;	//end-of-text flag
      }
      else
	line.nr++;			//line-number

      switch (line.start_p[0])		//first-character
      {
	case '#':			//comment?
	case '\n':			//line-feed?
	  break;			//nothing to do

	case '[':			//category-title?
	{
	  const pkg_STR_T* categoryName_p; //category-name pointer
	  pkg_STR_T category;		//category-name pointer

	  //check title for closing bracket and category-name
	  category.length = line.end_p - line.start_p - 2; //string-length
	  category.string_p = &line.start_p[1]; //string pointer

	  if (line.end_p[-1] != ']'	//no closing bracket?
	    || (categoryName_p = bsearch( //or binaray-search
	      &category,		//in: key pointer
	      pkg_puppyCategory_a,	//in: base buffer-pointer
	      GLB_ArrayLength_M(
		  pkg_puppyCategory_a),	//in: base-item-count
	      sizeof(pkg_STR_T),	//in: base-item-size
	      (pkg_COMPARE_FP_T) pkg_StringCompare)) //in: compare-function
					//out: item pointer
		== GLB_NULL)		//invalid item pointer?
	    ERR_ErrorFatalLine(		//error-fatal-line
	      (uchar*) "Invalid category title", //in: text pointer
	      convert_p->category.fileName_p, //in: file-name pointer
	      line.nr);			//in: line-number

	    //success: the category-index is the difference of the returned
	    //category pointer and the array address
	    package.iCategory = categoryName_p -
	      pkg_puppyCategory_a;	//category-index
	  break;
	}
	default:
	  //store all packages sorted to be able to find their category with
	  //a fast binary search
	  package.name.length = line.end_p - line.start_p; //string-length
	  package.name.string_p = line.start_p; //string pointer

	  pkg_SortedInsert(		//sorted-insert
	    sizeof(pkg_CTG_PACKAGE_T),	//in: item-size
	    &package,			//in: item pointer
	    (pkg_COMPARE_FP_T)
	      pkg_StringCompare,	//in: compare function-pointer
	    packageCount,		//in: list-item-count
	    package_bp);		//in/out: list pointer-pointer

	  if (++packageCount > itemMax) //package-count overflow?
	    //now has happend what is unlikely to happen: sigh and die
	    ERR_ErrorFatalLine(		//error-fatal-line
	      (uchar*) "To many small packages in the category file.\n"
	      "Insert some long comments.",
					//in: text pointer
	      convert_p->category.fileName_p, //in: file-name pointer
	      line.nr);			//in: line-number
	  break;
      }
    }
    while (line.eot_F == GLB_BOO_FALSE); //while not end-of-text

    convert_p->category.packageCount = packageCount; //package-count
  }
  memset(				//memory-set
    package.info_b,			//in/out: memory pointer
    0,					//in: byte
    sizeof(package.info_b));		//in: count

  excludeDebug_F = convert_p->excludeDebug_F; //exclude-debug flag
  line.nr = 0;				//line-number
  line.eot_F = GLB_BOO_FALSE;		//end-of-text flag
  line.start_p = (uchar*) text_bp;	//line-start pointer
  goto DCN4_L;				//skip the copying of the line-end

  //evaluate package-info-file, find first package
  do
  {
    //next line, skip line-feed
    line.start_p = &line.end_p[1];	//line-start pointer

    DCN4_L:
    //search line-end
    if ((line.end_p = (uchar*) strchr((char*) line.start_p, '\n'))
      == GLB_NULL)			//last-line?
    {
      line.end_p = (uchar*) &text_bp[textLength - 1]; //line-end pointer
      line.eot_F = GLB_BOO_TRUE;	//end-of-text flag
    }
    else
      line.nr++;			//line-number
  }
  while (pkg_DebianGetIdText(		//while debian-get-id-text
    dbc_debianId_a[IDI_PACKAGE].name,	//in: item-identifier
    line.start_p,			//in: line-start pointer
    line.end_p,				//in: line-end pointer
    &package.info_b[pkg_PKI_PACKAGE])	//in/out: text pointer
					//out: id-found
    == GLB_RSL_FAILURE			//failure
    && line.eot_F == GLB_BOO_FALSE);	//and not end-of-text

  //first package found, start conversion
  package.lineNr = line.nr;		//package-line-number
  convert_F = GLB_BOO_FALSE;		//convert flag
  iStart = 0;				//start-index

  while (line.eot_F == GLB_BOO_FALSE)	//while not end-of-text
  {
    uchar	i;			//index

    line.start_p = &line.end_p[1];	//line-start pointer

    if ((line.end_p = (uchar*) strchr((char*) line.start_p, '\n'))
      == GLB_NULL)			//last-line?
    {
      line.end_p = (uchar*) &text_bp[textLength - 1]; //line-end pointer
      line.eot_F = GLB_BOO_TRUE;	//end-of-text flag
      convert_F = GLB_BOO_TRUE;		//convert flag
    }
    else
      line.nr++;			//line-number

    //for all identifiers that have not been found yet
    for (i = iStart; i < GLB_ArrayLength_M(dbc_debianId_a); i++)
    {
      if (pkg_DebianGetIdText(		//while debian-get-id-text
	dbc_debianId_a[i].name,		//in: item-identifier
	line.start_p,			//in: line-start pointer
	line.end_p,			//in: line-end pointer
	&package.info_b[
	  dbc_debianId_a[i].iPuppy])	//in/out: package-info pointer
					//out: id-found
	== GLB_RSL_SUCCESS)		//success?
      {
	//do not compare already found ids: mark id found
	if (i == iStart)		//start-index
	  iStart++;			//start-index

	if (i == IDI_PACKAGE)		//package?
	{
	  //the next package id (Package:) was found, convert the current one
	  convert_F = GLB_BOO_TRUE;	//convert flag
	  break;
	}
      }
    }
    if (convert_F == GLB_BOO_TRUE)	//convert?
    {
      convert_F = GLB_BOO_FALSE;	//convert flag

      if (excludeDebug_F == GLB_BOO_FALSE //not exclude-debug?
	|| *((ulong*) &PACKAGE_P[
	  PACKAGE_LENGTH - (sizeof("-dbg") - 1)]) !=
	    GLB_BytesAsDWord_M('-', 'd', 'b', 'g')) //or no debug-extension?
	pkg_DebianConvertInfo(		//debian-convert-package-info
	  convert_p,			//in/out: convert pointer
	  &package);			//in: package pointer

      iStart = 0;			//start-index

      //continue with evaluation of the next package
      //delete previous package info except of next-package, which now
      //becomes the current package
      memset(				//memory-set
	package.info_b,			//in/out: memory pointer
	0,				//in: byte
	sizeof(package.info_b) - sizeof(pkg_STR_T)); //in: count

      package.info_b[pkg_PKI_PACKAGE] = package.info_b[pkg_PKI_NEXT_PACKAGE];

      //as the package info is assembled through all lines of a package and
      //the error handling shall not decrease performance, output the package
      //start line in case of an error
      package.lineNr = line.nr;		//package-line-number
    }
  }
  //conversion finished: flush text buffer rest to file
  if ((outputText_bp = convert_p->output.text_b) !=
      convert_p->output.line_p		//text in buffer?
    && write(				//and write
      convert_p->output.file_h,		//in: file handle
      outputText_bp,			//in: data buffer-pointer
      convert_p->output.line_p - outputText_bp) //in: data-length
					//out: bytes-written
      < 0)				//error?
      ERR_ErrorFatalSystem(		//error-fatal-system
	(uchar*) "Writing data base",	//in: text pointer
	convert_p->output.fileName_p);	//in: file-name pointer
}

//===========================================================================
static void pkg_DebianConvertInfo(	//debian-convert-package-info
  pkg_CONVERT_T* const convert_p,	//in: convert pointer
  const pkg_PACKAGE_T* const package_p)	//in: package pointer
{
  //definitions to make the function more readable
  #define CATEGORY_LENGTH\
    puppy_b[pkg_PKI_CATEGORY].length	//category-length
  #define CUT_BASE_LENGTH\
    puppy_b[pkg_PKI_CUT_BASE].length	//cut-base-length
  #define DEPS_LENGTH\
    puppy_b[pkg_PKI_DEPS].length	//dependencies-length
  #define DESCRIPTION_LENGTH\
    puppy_b[pkg_PKI_DESCRIPTION].length	//description-length
  #define DIRECTORY_LENGTH\
    puppy_b[pkg_PKI_DIRECTORY].length	//directory-length
  #define FILE_LENGTH\
    puppy_b[pkg_PKI_FILE].length	//file-length
  #define RELEASE_LENGTH\
    puppy_b[pkg_PKI_RELEASE].length	//release-length
  #define SIZE_LENGTH\
    puppy_b[pkg_PKI_SIZE].length	//size-length
  #define VERSION_LENGTH\
    puppy_b[pkg_PKI_VERSION].length	//version-length

  #define CATEGORY_P puppy_b[pkg_PKI_CATEGORY].string_p //category pointer
  #define CUT_BASE_P puppy_b[pkg_PKI_CUT_BASE].string_p //cut-base pointer
  #define DEPS_P puppy_b[pkg_PKI_DEPS].string_p //dependencies pointer
  #define DESCRIPTION_P\
    puppy_b[pkg_PKI_DESCRIPTION].string_p //description pointer
  #define DIRECTORY_P puppy_b[pkg_PKI_DIRECTORY].string_p //directory pointer
  #define FILE_P puppy_b[pkg_PKI_FILE].string_p //file pointer
  #define RELEASE_P puppy_b[pkg_PKI_RELEASE].string_p //release pointer
  #define SIZE_P puppy_b[pkg_PKI_SIZE].string_p //size pointer
  #define VERSION_P puppy_b[pkg_PKI_VERSION].string_p //version pointer

  MEM_SIZE_T	fileNameLength;		//file-name-length
  MEM_SIZE_T	i;			//index
  MEM_SIZE_T	iCategory;		//category-index
  pkg_CTG_PACKAGE_T* categoryPackage_p;	//category-package pointer
  pkg_STR_T	puppy_b[pkg_PKI_COUNT];	//puppy-package-info buffer
  uchar*	row_p;			//row pointer
  uchar*	char_p;			//character pointer
  uchar*	end_p;			//insertion-end pointer

  #ifdef pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
  MEM_SIZE_T	lineLength;		//line-length

  #else //pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
  uchar*	rowLimit_p;		//row-limit pointer
  #endif //pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer

  //copy debian-package-info to puppy-package-info
  memcpy(				//memory-copy
    puppy_b,				//in/out: target pointer
    package_p->info_b,			//in: source pointer
    sizeof(puppy_b));			//in: count

  fileNameLength = DIRECTORY_LENGTH;	//file-name-length

  //extract file from file-name which is found at the directory index
  if ((FILE_P = memrchr(		//memory-reverse-character
    DIRECTORY_P,			//in: memory pointer
    '/',				//in: character
    fileNameLength))			//in: count
					//out: character pointer
    == GLB_NULL)			//invalid character pointer?
    FILE_P = DIRECTORY_P;		//file pointer
  else
  {
    //skip delimiter
    FILE_P++;				//file pointer
    DIRECTORY_LENGTH--;			//directory-length
  }
  //the cut base name also points to the file, it is only shorter
  CUT_BASE_P = FILE_P;			//cut-base pointer
  CUT_BASE_LENGTH =			//cut-base-length
    FILE_LENGTH = &DIRECTORY_P[fileNameLength] - FILE_P; //file-length

  //cut the directory length
  DIRECTORY_LENGTH -= FILE_LENGTH;	//directory-length

  //cut the cut-base-length
  if ((char_p = memrchr(		//memory-reverse-character
    CUT_BASE_P,				//in: memory pointer
    '.',				//in: character
    CUT_BASE_LENGTH))			//in: count
					//out: character pointer
    != GLB_NULL)			//valid character pointer?
  {
    CUT_BASE_LENGTH = char_p - CUT_BASE_P; //cut-base-length

  if ((char_p = memrchr(		//memory-reverse-character
      CUT_BASE_P,			//in: memory pointer
      '-',				//in: character
      CUT_BASE_LENGTH))			//in: count
					//out: character pointer
      != GLB_NULL)			//valid character pointer?
      CUT_BASE_LENGTH = char_p - CUT_BASE_P; //cut-base-length
  }
  //build version release
  if ((char_p = memrchr(		//memory-reverse-character
    VERSION_P,				//in: memory pointer
    '-',				//in: character
    VERSION_LENGTH))			//in: count
					//out: character pointer
    != GLB_NULL)			//valid character pointer?
  {
    RELEASE_P = &char_p[1];		//release pointer
    RELEASE_LENGTH = &VERSION_P[
      VERSION_LENGTH] - RELEASE_P;	//release-length
    VERSION_LENGTH -= RELEASE_LENGTH + 1; //version-length
  }
  //cut possible epoch from version
  if ((char_p = memchr(			//memory-character
    VERSION_P,				//in: memory pointer
    ':',				//in: character
    VERSION_LENGTH))			//in: count
					//out: character pointer
    != GLB_NULL)			//valid character pointer?
  {
    //skip colon
    char_p++;				//character pointer
    VERSION_LENGTH -= char_p - VERSION_P; //version-length
    VERSION_P = char_p;			//version pointer
  }
  //build category via a binary search in the category-package buffer
  //use BuildingBlock if category was not found
  iCategory = (categoryPackage_p = bsearch( //binaray-search
    &puppy_b[pkg_PKI_PACKAGE],		//in: key pointer
    convert_p->category.package_bp,	//in: base buffer-pointer
    convert_p->category.packageCount,	//in: base-item-count
    sizeof(pkg_CTG_PACKAGE_T),		//in: base-item-size
    (pkg_COMPARE_FP_T) pkg_StringCompare)) //in: compare-function
					//out: item pointer
    != GLB_NULL?			//valid item pointer?
    categoryPackage_p->iCategory:	//category-index
    pkg_ICT_BUILDINGBLOCK;		//building-block

  puppy_b[pkg_PKI_CATEGORY] = pkg_puppyCategory_a[iCategory]; //category

  #ifndef pkg_POLITICAL_CORRECT		//avoid dubious fast constructs
  //debian outputs the size in KBytes leaving away the K, Puppy needs it
  //Ugly but fast: As the read text is not used any more, write the K to the
  //text and increase the size length by one

  SIZE_P[SIZE_LENGTH] = 'K';		//size pointer
  SIZE_LENGTH++;			//size-length
  #endif //pkg_POLITICAL_CORRECT	//avoid dubious fast constructs

  row_p =				//row pointer
    end_p = convert_p->output.line_p;	//insertion-end pointer

  #ifdef pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
  lineLength = 0;			//line-length

  #else //pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
  rowLimit_p = end_p + pkg_LINE_MAX - 4; //row-limit pointer
  #endif //pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer

  for (i = 0; i < pkg_PKI_DEPS; i++)	//for row until dependencies
  {
    MEM_SIZE_T	itemLength;		//item-length

    itemLength = puppy_b[i].length;	//item-length

    #ifdef pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
    if ((lineLength += itemLength + 1) >=
      pkg_LINE_MAX - 4)			//line-length overflow?

    #else //pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer
    if (end_p + itemLength >= rowLimit_p) //row-length overflow?
    #endif //pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer

      goto DCI6_L;			//error database row to long

    end_p = MEM_Copy(			//memory-copy
      itemLength,			//in: count
      puppy_b[i].string_p,		//in: source pointer
      end_p);				//in/out: target pointer
					//out: target-end pointer

    #ifdef pkg_POLITICAL_CORRECT	//avoid dubious fast constructs
    //as this loop is run several thousand times in big package file
    //every unnecessary query decreases performance
    if (i == pkg_PKI_SIZE)		//size?
    {
      *end_p++ = 'K';			//kilo-bytes

      #ifdef pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer
      lineLength++;			//line-length
      #endif //pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer
    }
    #endif //pkg_POLITICAL_CORRECT	//avoid dubious fast constructs

    *end_p++ = pkg_SEPARATOR;		//separator
  }
  //as the puppy dependencies never will be longer than the debian
  //dependencies, continue the size check but do not copy yet
  for (; i < pkg_PKI_ROW_COUNT; i++)	//for row rest
  {
    #ifdef pkg_LOOP_WITH_LENGTH		//loop with length instead of pointer
    if ((lineLength += puppy_b[i].length + 1) >=
      pkg_LINE_MAX - 4)			//line-length overflow?

    #else //pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer
    if (end_p + puppy_b[i].length >= rowLimit_p) //row-length overflow?
    #endif //pkg_LOOP_WITH_LENGTH	//loop with length instead of pointer

      DCI6_L:
      ERR_ErrorFatalLine(		//error-fatal-line
	(uchar*) "Puppy database row to becomes to long", //in: text pointer
	convert_p->inputFileName_p,	//in: file-name pointer
	package_p->lineNr);		//in: line-number
  }
  //dependencies and description will fit: build and copy dependencies
  if (DEPS_LENGTH != 0)			//valid dependency-length?
  {
    MEM_SIZE_T	restLength;		//rest-length
    uchar*	dep_p;			//dependency pointer
    uchar*	found_p;		//found pointer

    dep_p = DEPS_P;			//dependency pointer

    for (restLength = DEPS_LENGTH; ; )	//for all dendencies
    {
      MEM_SIZE_T oldRestLength;		//old-rest-length
      MEM_SIZE_T depLength;		//dependency-length

      oldRestLength = restLength;	//old-rest-length

      //search for comma, space or pipe
      if ((found_p = MEM_CharFromGroup(	//memory-character-from-group
	(uchar*) ", |",			//in: group pointer
	restLength,			//in: count
	dep_p))				//in: memory pointer
					//out: character pointer
	== GLB_NULL)			//invalid character pointer?
      {
	//no comma, space or pipe found: last dependency
	depLength = restLength;		//dependency-length
	restLength = 0;			//rest-length
      }
      else
      {
	//comma, space or pipe found
	uchar	chr;			//character

	depLength = found_p - dep_p;	//dependency-length
	restLength -= depLength + 1;	//rest-length
	chr = *found_p++;		//character, skip it

	if (chr == ' '			//space?
	  || chr == '|')		//or pipe?
	{
	  //skip everything between space/pipe and comma
	  uchar* commaFound_p;		//comma-found pointer

	  if ((commaFound_p = memchr(	//memory-character
	    found_p,			//in: memory pointer
	    ',',			//in: character
	    restLength))		//in: count
					//out: character pointer
	    != GLB_NULL)		//valid character pointer?
	  {
	    commaFound_p++;		//skip comma
	    restLength -= commaFound_p - found_p; //rest-length
	    found_p = commaFound_p;	//found pointer
	  }
	  else
	    //no more comma: nothing more to do
	    restLength = 0;		//rest-length
	}
	while (isspace(*found_p++)
	  && restLength != 0)		//skip spaces
	  restLength--;			//rest-length
      }
      //output dependency in puppy format
      *end_p++ = '+';			//end pointer

      end_p = MEM_Copy(			//memory-copy
	depLength,			//in: count
	dep_p,				//in: source pointer
	end_p);				//in/out: target pointer
					//out: target-end pointer

      if (restLength == 0)		//no more rest-length?
	break;				//finish loop

      *end_p++ = ',';			//end pointer
      dep_p += oldRestLength - restLength; //dependency pointer
    }
  }
  *end_p++ = pkg_SEPARATOR;		//separator

  //copy description
  end_p = MEM_Copy(			//memory-copy
    DESCRIPTION_LENGTH,			//in: count
    DESCRIPTION_P,			//in: source pointer
    end_p);				//in/out: target pointer
					//out: target-end pointer

  *((ushort*) end_p) =
    GLB_BytesAsWord_M(pkg_SEPARATOR, '\n'); //separator and line-feed
  end_p += 2;				//end pointer
  //*end_p = '\0';			//end-of-string for debugging

  if (end_p - convert_p->output.text_b <
    pkg_LINE_MAX)			//no more space for 1 more line left?
  {
    //flush text to file
    if (write(				//write
      convert_p->output.file_h,		//in: file handle
      row_p,				//in: data buffer-pointer
      end_p - row_p)			//in: data-length
					//out: bytes-written
      < 0)				//error?
      ERR_ErrorFatalSystem(		//error-fatal-system
	(uchar*) "Writing data base",	//in: text pointer
	convert_p->output.fileName_p);	//in: file-name pointer

    //reset line pointer to text buffer start
    convert_p->output.line_p =
      convert_p->output.text_b;		//output-line pointer
  }
  else
    convert_p->output.line_p = end_p;	//output-line pointer
}

//===========================================================================
static GLB_RESULT_T pkg_DebianGetIdText( //debian-get-id-text
  const pkg_STR_T	itemId,		//in: item-identifier
  const uchar* const	lineStart_p,	//in: line-start pointer
  const uchar* const	lineEnd_p,	//in: line-end pointer
  pkg_STR_T* const	text_p)		//in/out: text pointer
					//out: id-found
//check line for a valid identifier, if found mark the text following the id
//return whether an id was found or not
{
  const uchar*	textEnd_p;		//text-end pointer
  const uchar*	textStart_p;		//text-start pointer

  if (memcmp(				//memory-compare
    lineStart_p,			//in: memory-1 pointer
    itemId.string_p,			//in: memory-1 pointer
    itemId.length)			//in: count
					//out: compare-result
    != 0)				//no item found?
    return GLB_RSL_FAILURE;		//failure

  textStart_p = &lineStart_p[itemId.length]; //text-start pointer
  textEnd_p = lineEnd_p;		//text-end pointer

  while (textStart_p != textEnd_p	//skip spaces at text start
    && isspace(*textStart_p))
    textStart_p++;			//text-start pointer

  while (textEnd_p != textStart_p	//cut spaces at text end
    && isspace(textEnd_p[-1]))
    textEnd_p--;			//text-end pointer

  text_p->length = textEnd_p - textStart_p; //string-length
  text_p->string_p = (uchar*) textStart_p; //string pointer
  return GLB_RSL_SUCCESS;		//success
}

//===========================================================================
static int pkg_StringCompare(		//string-compare
  const pkg_STR_T* const key_p,		//in: key pointer
  const pkg_STR_T* const listItem_p)	//in: list-item pointer
					//out: comparision-result
//compare length-strings
{
  int		result;			//result

  if ((result = memcmp(			//memory-compare
    key_p->string_p,			//in: memory-1 pointer
    listItem_p->string_p,		//in: memory-2 pointer
    GLB_Min_M(				//  minimum
      listItem_p->length,		//  in: value-1
      key_p->length)))			//  in: value-2
					//  out: minimum
					//in: count
    == 0)				//same?
    //return shorter string before longer one
    return key_p->length - listItem_p->length; //result

  return result;			//result
}

//===========================================================================
void pkg_SortedInsert(			//sorted-insert
  const MEM_SIZE_T	itemSize,	//in: item-size
  const void* const	item_p,		//in: item pointer
  const pkg_COMPARE_FP_T compare_fp,	//in: compare function-pointer
  const MEM_SIZE_T	itemCount,	//in: list-item-count
  void*			list_p)		//in/out: list pointer
//insert an item into a sorted list
{
  #define LIST_P ((uchar*) list_p)	//list pointer

  MEM_SIZE_T	low;			//low
  uchar*	insert_p;		//insert pointer

  low = 0;				//low

  if (itemCount != 0)			//valid item-count?
  {
    MEM_SIZE_T	high;			//high

    high = (MEM_SIZE_T) (itemCount - 1); //high

    while (low <= high)			//for whole search
    {
      int	compareResult;		//compare-result
      MEM_SIZE_T i;			//index

      i = (MEM_SIZE_T) ((low + high) >> 1); //index

      compareResult = compare_fp(	//compare function-pointer
	&LIST_P[i * itemSize],		//in: item-1 pointer
	item_p);			//in: item-2 pointer
					//out compare-result

      if (compareResult < 0)		//compare-result smaller?
	low = (MEM_SIZE_T) (i + 1); 	//low
      else
      {
	if (compareResult == 0)		//compare-result equal?
	  low = i;			//low

	if (i == 0)			//nothing more to search?
	  break;

	high = (MEM_SIZE_T) (i - 1); 	//high
      }
    }
  }
  insert_p = &LIST_P[low * itemSize];	//insert pointer

  //move items up
  memmove(				//memory-move
    &insert_p[itemSize],		//in/out: target pointer
    insert_p,				//in: source pointer
    (itemCount - low) * itemSize);	//in: count

  //insert sorted item
  memcpy(				//memory-copy
    insert_p,				//in/out: target pointer
    item_p,				//in: source pointer
    itemSize);				//in: count
}

//===========================================================================
static void pkg_Usage(void)		//usage
{
  printf(
  "Syntax: ppkg OPTIONS\n"
  "Puppy Linux Package Handler (under construction) Version 0.0.0\n"
  "\n"
  "OPTIONS:\n"
  "-c DISTRIB [-i CATEGORY-FILE] [-x KIND] DIST.-FILE [PUPPY-FILE]\n"
  "   convert distribution-package-info-file to puppy-package-database-file\n"
  "   DISTRIB: arch debian slackware ubuntu\n"
  "   -i CATEGORY-FILE: include puppy-category-file to set menu categories\n"
  "   -x KIND: exclude package-kinds\n"
  "      KIND: debug\n"
  "   if PUPPY-FILE is omitted use standard output\n"
  "\n"
  "-? Display this help\n\n");

  exit(1);				//exit with error
}

//===========================================================================
//GLOBAL FUNCTIONS
//===========================================================================

//===========================================================================
int main(				//main-function
  const int		argCount,	//in: argument-count
  const char* const* const arg_pp)	//in: argument pointer-pointer
{
  typedef enum
  {
    DST_ARCH		= 0,		//arch
    DST_DEBIAN		= 1,		//debian
    DST_SLACKWARE	= 2,		//slackware
    DST_UBUNTU		= 3,		//ubuntu
    DST_COUNT		= 4,		//count
    DST_INVALID		= 0xFF,		//invalid
  } DISTRIB_ET;				//distribution enum-type

  static const uchar* const mai_distribName_sa[] =
					//distribution-name string-array
  {
    (uchar*) "arch",
    (uchar*) "debian",
    (uchar*) "slackware",
    (uchar*) "ubuntu",
  };

  const uchar* const* currentArg_pp;	//current-argument pointer-pointer
  const uchar* const* distribItem_pp;	//distribution-item pointer-pointer
  const uchar*	distribution_p;		//distribution pointer
  const uchar*	exclude_p;		//exclude pointer
  const uchar*	option_p;		//option pointer
  DISTRIB_ET	distribKind;		//distribution-kind
  MEM_SIZE_T	textLength;		//text-length
  pkg_CONVERT_T	convert;		//convert
  uchar*	text_bp;		//text buffer-pointer

  memset(				//memory-set
    &convert,				//in/out: memory pointer
    0,					//in: byte
    sizeof(convert));			//in: count

  distribution_p = GLB_NULL;		//distribution pointer
  exclude_p = GLB_NULL;			//exclude pointer
  currentArg_pp = (const uchar* const*) arg_pp;
					//current-argument pointer-pointer

  ERR_Init(				//error-initialisation
    currentArg_pp[0]);			//in: program-name pointer

  //evaluate command line
  while ((option_p = *++currentArg_pp) != GLB_NULL) //while argument is valid
  {
    switch (*(ushort*) option_p)	//first two characters in argument
    {
      const uchar** optionString_pp;	//option-string pointer-pointer

      case GLB_BytesAsWord_M('-', 'c'): //convert?
	optionString_pp = &distribution_p; //option-string pointer-pointer
	goto MAI2_L;			//evaluate option-string

      case GLB_BytesAsWord_M('-', 'i'): //include?
	optionString_pp =
	  &convert.category.fileName_p;	//option-string pointer-pointer
	goto MAI2_L;			//evaluate option-string

      case GLB_BytesAsWord_M('-', 'x'): //exclude?
	optionString_pp = &exclude_p;	//option-string pointer-pointer

	MAI2_L:
	//option-string right behind option?
	if (option_p[2] != '\0')	//not end-of-string?
	  *optionString_pp = &option_p[2]; //include-file-name pointer
	else
	{
	  if ((option_p = *++currentArg_pp)
	      == GLB_NULL		//next argument invalid?
	    || option_p[0] == '-')	//or option?
	    pkg_Usage();		//usage

	  //option-string in next argument and next argument no option
	  *optionString_pp = option_p;	//option-string pointer-pointer
	}
	break;

      default:
	if (option_p[0] == '-')		//invalid option?
	  pkg_Usage();			//usage

	if (convert.inputFileName_p ==
	  GLB_NULL)			//invalid input-file-name pointer?
	  convert.inputFileName_p = option_p; //input-file-name pointer
	else if (convert.output.fileName_p ==
	  GLB_NULL)			//invalid output-file-name pointer?
	  convert.output.fileName_p = option_p; //output-file-name pointer
	else				//more than two file-names?
	  pkg_Usage();			//usage
	break;
    }
  }
  if (distribution_p == GLB_NULL	//invalid distribution pointer?
    || convert.inputFileName_p ==
      GLB_NULL)				//or invalid input-file-name pointer?
    pkg_Usage();			//usage

  //get distribution kind
  distribKind = (distribItem_pp = bsearch( //binaray-search
    distribution_p,			//in: key pointer
    mai_distribName_sa,			//in: base buffer-pointer
    GLB_ArrayLength_M(mai_distribName_sa), //in: base-item-count
    sizeof(uchar*),			//in: base-item-size
    (pkg_COMPARE_FP_T) pkg_CStringItemCompare)) //in: compare-function
					//out: item pointer
    != GLB_NULL?			//valid item pointer?
    distribItem_pp - mai_distribName_sa: //array-index
    DST_INVALID;			//invalid

  switch (distribKind)			//distribution-kind
  {
    case DST_ARCH:			//arch?
    case DST_SLACKWARE:			//slackware?
    {
      uchar	text_b[80];		//text buffer

      sprintf((char*) text_b, "Conversion of %s not implemented yet",
	distribution_p);

      ERR_ErrorFatal(			//error-fatal
	text_b);			//in: text pointer
      break;
    }
    default:				//invalid?
      ERR_ErrorFatal(			//error-fatal
	(uchar*) "Invalid distribution name"); //in: text pointer
					//fall through!!!

    case DST_DEBIAN:			//debian?
    case DST_UBUNTU:			//ubunt?u
      break;
  }
  //read the whole file to memory, linux has enough, do everything with
  //pointers to become as fast as possible
  FIO_FileToMemory(			//file-to-memory
    convert.inputFileName_p,		//in: file-name pointer
    FIO_OPT_EOS_0 | FIO_OPT_OUTPUT_ERR | FIO_OPT_EXIT_ON_ERR, //in: option
    &text_bp,				//in/out: memory pointer-pointer
    &textLength);			//in/out: memory-size pointer
					//out: file-result

  if (convert.output.fileName_p ==
    GLB_NULL)				//invalid output-file-name pointer?
  {
    //output to stdout
    convert.output.fileName_p = (uchar*) "stdout"; //output-file-name pointer
    convert.output.file_h = 1;		//output-file handle
  }
  else
  {
    //open output file
    if ((convert.output.file_h = open(	//open
      (char*) convert.output.fileName_p, //in: file-name pointer
      O_CREAT | O_WRONLY | O_TRUNC))	//in: openKind
					//out: file handle
      <= 0)				//failure?
      ERR_ErrorFatalSystem(		//error-fatal-system
	(uchar*) "Creating output file", //in: text pointer
	convert.output.fileName_p);	//in: file-name pointer
  }
  convert.excludeDebug_F = (GLB_BOOL_T)	//exclude-debug flag
    (exclude_p != GLB_NULL		//valid exclude pointer?
      && strcmp((char*) exclude_p, "debug") == 0);

  pkg_DebianConvert(			//debian-convert
    &convert,				//in/out: convert pointer
    textLength,				//in: text-length
    text_bp);				//in: text buffer-pointer

  close(				//file-close
    convert.output.file_h);		//in: file handle
					//out: close-result

  free(					//free-memory
    text_bp);				//in: memory pointer

  return 0;				//exit with success
}
